/*
 * Licensed under a Creative Commons Attribution 2.5 Slovenia License
 * http://creativecommons.org/licenses/by/2.5/si/
 * 2009 TineL Studio
 */

package net.tinelstudio.gis.routing.service;

import java.util.concurrent.TimeoutException;

import net.tinelstudio.gis.common.NotEnabledException;
import net.tinelstudio.gis.common.ServiceException;
import net.tinelstudio.gis.routing.dto.Route;
import net.tinelstudio.gis.routing.searchalgorithm.astar.heuristic.Heuristic;
import net.tinelstudio.gis.routing.searchalgorithm.cost.Cost;

import com.vividsolutions.jts.geom.Coordinate;

/**
 * Defines the client side of the Routing service.
 * 
 * @author TineL
 */
public interface RoutingService {

  /**
   * Finds a route with the default {@link Heuristic} and {@link Cost}
   * functions.
   * 
   * @param startLocation the start location of the route (<code>null</code> not
   *        permitted)
   * @param goalLocation the end location of the route (<code>null</code> not
   *        permitted)
   * @return a found route or <code>null</code> if none found
   * @throws NotEnabledException If the service is not enabled at the moment
   * @throws TimeoutException If the routing search is taking too much time to
   *         find a route
   * @throws ServiceException If some internal service or server error occurs
   */
  Route findRoute(Coordinate startLocation, Coordinate goalLocation)
    throws NotEnabledException, TimeoutException, ServiceException;

  /**
   * Finds a route.
   * 
   * @param startLocation the start location of the route (<code>null</code> not
   *        permitted)
   * @param goalLocation the end location of the route (<code>null</code> not
   *        permitted)
   * @param cost the cost function g(n) (<code>null</code> not permitted)
   * @param heuristic the heuristic function h(n) (<code>null</code> not
   *        permitted)
   * @return a found route or <code>null</code> if none found
   * @throws NotEnabledException If the service is not enabled at the moment
   * @throws TimeoutException If the routing search is taking too much time to
   *         find a route
   * @throws ServiceException If some internal service or server error occurs
   */
  Route findRoute(Coordinate startLocation, Coordinate goalLocation, Cost cost,
    Heuristic heuristic) throws NotEnabledException, TimeoutException,
    ServiceException;

  /**
   * @return the service timeout of the routing search in milliseconds. If <= 0,
   *         no timeout is defined.
   * @throws NotEnabledException If the service is not enabled at the moment
   */
  int getTimeout() throws NotEnabledException;

  /**
   * @return the current server time in milliseconds
   */
  long getServerTime();
}
