
package net.tinelstudio.gis.routing.searchalgorithm.astar;

import static org.easymock.EasyMock.expect;
import static org.easymock.classextension.EasyMock.createMock;
import static org.easymock.classextension.EasyMock.replay;
import static org.easymock.classextension.EasyMock.verify;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertSame;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeoutException;

import net.tinelstudio.gis.model.dao.FindingDao;
import net.tinelstudio.gis.model.domain.Street;
import net.tinelstudio.gis.model.domain.StreetNode;
import net.tinelstudio.gis.routing.dto.Route;
import net.tinelstudio.gis.routing.searchalgorithm.astar.DefaultAStarSearchAlgorithm.StreetNodeDecorator;
import net.tinelstudio.gis.routing.searchalgorithm.astar.heuristic.HeuristicFunction;
import net.tinelstudio.gis.routing.searchalgorithm.cost.CostFunction;

import org.easymock.IAnswer;
import org.junit.Test;

import com.vividsolutions.jts.geom.Point;

/**
 * @author TineL
 */
public class DefaultAStarSearchAlgorithmTest extends
  AbstractAStarSearchAlgorithmTest {

  @Override
  protected DefaultAStarSearchAlgorithm createInstance() {
    return new DefaultAStarSearchAlgorithm();
  }

  protected DefaultAStarSearchAlgorithm createInstance(final Point startPoint) {
    return new DefaultAStarSearchAlgorithm() {
      @Override
      protected double h(Point pos) {
        assertSame(startPoint, pos);
        return 15;
      }
    };
  }

  private void testRun(final int timeout) throws TimeoutException {
    // Init
    final Point startPoint=createMock(Point.class);

    StreetNode startNode=createMock(StreetNode.class);
    expect(startNode.getId()).andReturn(1L).anyTimes();
    expect(startNode.getPoint()).andReturn(startPoint).anyTimes();

    Point goalPoint=createMock(Point.class);

    StreetNode goalNode=createMock(StreetNode.class);
    expect(goalNode.getId()).andReturn(10L).anyTimes();
    expect(goalNode.getPoint()).andReturn(goalPoint).anyTimes();

    CostFunction costFunction=createMock(CostFunction.class);

    HeuristicFunction heuristicFunction=createMock(HeuristicFunction.class);

    FindingDao findingDao=createMock(FindingDao.class);
    expect(findingDao.findConnectedStreets(startNode)).andAnswer(
      new IAnswer<List<Street>>() {
        @Override
        public List<Street> answer() throws Throwable {
          // Add 500 ms to eliminate System.currentMillis() small inaccuracy
          Thread.sleep(500+timeout);
          return Collections.emptyList();
        }
      });

    replay(startNode, goalNode, findingDao);

    // Run
    DefaultAStarSearchAlgorithm aStar=createInstance(startPoint);
    aStar.setStartNode(startNode);
    aStar.setGoalNode(goalNode);
    aStar.setCostFunction(costFunction);
    aStar.setHeuristicFunction(heuristicFunction);
    aStar.setFindingDao(findingDao);
    aStar.setTimeout(timeout);
    Route route=aStar.run();

    // Verify
    verify(startNode, goalNode, findingDao);
    assertNull(route);
  }

  @Test
  public void testRun() throws TimeoutException {
    testRun(0); // Unlimited timeout
  }

  @Test(expected=TimeoutException.class)
  public void testRunTimeout() throws TimeoutException {
    testRun(1); // Timeout only 1 ms
  }

  @Test
  public void testReconstructPath() {
    // Init
    Long node1Id=10L;
    StreetNodeDecorator node1=createMock(StreetNodeDecorator.class);
    expect(node1.getId()).andReturn(node1Id).anyTimes();

    Long node2Id=11L;
    StreetNodeDecorator node2=createMock(StreetNodeDecorator.class);
    expect(node2.getId()).andReturn(node2Id).anyTimes();

    Long node3Id=12L;
    StreetNodeDecorator node3=createMock(StreetNodeDecorator.class);
    expect(node3.getId()).andReturn(node3Id).anyTimes();

    Long goalNodeId=100L;
    StreetNodeDecorator goalNode=createMock(StreetNodeDecorator.class);
    expect(goalNode.getId()).andReturn(goalNodeId).anyTimes();

    Map<Long, StreetNodeDecorator> nodeMap=new HashMap<Long, StreetNodeDecorator>();
    nodeMap.put(goalNodeId, node1);
    nodeMap.put(node1Id, node2);
    nodeMap.put(node2Id, node3);
    nodeMap.put(node3Id, null);

    Street street1=createMock(Street.class);
    Street street2=createMock(Street.class);
    Street street3=createMock(Street.class);
    Street goalStreet=createMock(Street.class);

    Map<Long, Street> streetMap=new HashMap<Long, Street>();
    streetMap.put(goalNodeId, goalStreet);
    streetMap.put(node1Id, street1);
    streetMap.put(node2Id, street2);
    streetMap.put(node3Id, street3);

    replay(node1, node2, node3, goalNode);

    // Run
    DefaultAStarSearchAlgorithm aStar=createInstance();
    List<Street> streets=aStar.reconstructPath(nodeMap, streetMap, goalNode);

    // Verify
    verify(node1, node2, node3, goalNode);
    assertSame(street3, streets.get(0));
    assertSame(street2, streets.get(1));
    assertSame(street1, streets.get(2));
    assertSame(goalStreet, streets.get(3));
  }
}
