/*
 * Licensed under a Creative Commons Attribution 2.5 Slovenia License
 * http://creativecommons.org/licenses/by/2.5/si/
 * 2009 TineL Studio
 */

package net.tinelstudio.gis.routing.service;

import static org.easymock.EasyMock.expect;
import static org.easymock.classextension.EasyMock.createMock;
import static org.easymock.classextension.EasyMock.replay;
import static org.easymock.classextension.EasyMock.verify;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import java.util.concurrent.TimeoutException;

import net.tinelstudio.gis.common.NotEnabledException;
import net.tinelstudio.gis.common.service.AbstractEnablableServiceTest;
import net.tinelstudio.gis.model.dao.FindingDao;
import net.tinelstudio.gis.model.domain.StreetNode;
import net.tinelstudio.gis.routing.dto.DefaultRoute;
import net.tinelstudio.gis.routing.dto.Route;
import net.tinelstudio.gis.routing.searchalgorithm.astar.DefaultAStarSearchAlgorithm;
import net.tinelstudio.gis.routing.searchalgorithm.astar.heuristic.Heuristic;
import net.tinelstudio.gis.routing.searchalgorithm.astar.heuristic.HeuristicFunction;
import net.tinelstudio.gis.routing.searchalgorithm.astar.heuristic.HeuristicFunctionFactory;
import net.tinelstudio.gis.routing.searchalgorithm.cost.Cost;
import net.tinelstudio.gis.routing.searchalgorithm.cost.CostFunction;
import net.tinelstudio.gis.routing.searchalgorithm.cost.CostFunctionFactory;

import org.junit.Test;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.Point;

/**
 * @author TineL
 */
public class RoutingServiceImplTest extends AbstractEnablableServiceTest {

  @Override
  protected RoutingServiceImpl createInstance() {
    return new RoutingServiceImpl();
  }

  protected RoutingServiceImpl createInstance(
    final Coordinate expectedStartLocation,
    final Coordinate expectedGoalLocation,
    final CostFunction expectedCostFunction,
    final HeuristicFunction expectedHeuristicFunction, final Route returnRoute) {
    return new RoutingServiceImpl() {
      @Override
      protected Route findRoute(Coordinate startLocation,
        Coordinate goalLocation, CostFunction costFunction,
        HeuristicFunction heuristicFunction) throws TimeoutException {
        assertSame(expectedStartLocation, startLocation);
        assertSame(expectedGoalLocation, goalLocation);
        assertSame(expectedCostFunction, costFunction);
        assertSame(expectedHeuristicFunction, heuristicFunction);
        return returnRoute;
      }
    };
  }

  protected RoutingServiceImpl createInstance(
    final Coordinate expectedLocation1, final Coordinate expectedLocation2,
    final StreetNode returnNode,
    final DefaultAStarSearchAlgorithm returnSearchAlgorithm) {
    return new RoutingServiceImpl() {
      @Override
      protected StreetNode findNearestStreetNode(Coordinate c1) {
        assertTrue(expectedLocation1==c1||expectedLocation2==c1);
        return returnNode;
      }

      @Override
      protected DefaultAStarSearchAlgorithm createSearchAlgorithm() {
        return returnSearchAlgorithm;
      }
    };
  }

  @Test
  public void testSetCostFunctionFactory() {
    CostFunctionFactory factory=createMock(CostFunctionFactory.class);
    RoutingServiceImpl service=createInstance();
    service.setCostFunctionFactory(factory);
    assertSame(factory, service.getCostFunctionFactory());
  }

  @Test
  public void testSetHeuristicFunctionFactory() {
    HeuristicFunctionFactory factory=createMock(HeuristicFunctionFactory.class);
    RoutingServiceImpl service=createInstance();
    service.setHeuristicFunctionFactory(factory);
    assertSame(factory, service.getHeuristicFunctionFactory());
  }

  @Test
  public void testSetFindingDao() {
    FindingDao findingDao=createMock(FindingDao.class);
    RoutingServiceImpl service=createInstance();
    service.setFindingDao(findingDao);
    assertSame(findingDao, service.getFindingDao());
  }

  @Test
  public void testSetGeometryFactory() {
    GeometryFactory gf=createMock(GeometryFactory.class);
    RoutingServiceImpl service=createInstance();
    service.setGeometryFactory(gf);
    assertSame(gf, service.getGeometryFactory());
  }

  @Test(expected=NotEnabledException.class)
  public void testGetTimeoutDisabled() throws Exception {
    RoutingServiceImpl service=createInstance();
    service.setEnabled(false);
    service.getTimeout();
  }

  @Test
  public void testSetTimeout() throws Exception {
    int timeout=(int)(Math.random()*1000-100);
    RoutingServiceImpl service=createInstance();
    service.setTimeout(timeout);
    assertEquals(timeout, service.getTimeout());
  }

  @Test
  public void testSetStreetNodeMaxDistanceMeters() {
    int streetNodeMaxDistanceMeters=(int)(Math.random()*1000);
    RoutingServiceImpl service=createInstance();
    service.setStreetNodeMaxDistanceMeters(streetNodeMaxDistanceMeters);
    assertEquals(streetNodeMaxDistanceMeters, service
      .getStreetNodeMaxDistanceMeters());
  }

  @Test
  public void testGetServerTime() {
    double time=System.currentTimeMillis();
    RoutingServiceImpl service=createInstance();
    assertEquals(time, service.getServerTime(), 1000); // 1000 ms delta
  }

  @Test(expected=NotEnabledException.class)
  public void testFindRoute1Disabled() throws Exception {
    RoutingServiceImpl service=createInstance();
    service.setEnabled(false);
    service.findRoute(null, null, (Cost)null, (Heuristic)null);
  }

  @Test
  public void testFindRoute1() throws Exception {
    // Init
    Coordinate startLocation=createMock(Coordinate.class);
    Coordinate goalLocation=createMock(Coordinate.class);

    Cost cost=createMock(Cost.class);
    CostFunction costFunction=createMock(CostFunction.class);
    CostFunctionFactory costFunctionFactory=createMock(CostFunctionFactory.class);
    expect(costFunctionFactory.createCostFunction(cost))
      .andReturn(costFunction);

    Heuristic heuristic=createMock(Heuristic.class);
    HeuristicFunction heuristicFunction=createMock(HeuristicFunction.class);
    HeuristicFunctionFactory heuristicFunctionFactory=createMock(HeuristicFunctionFactory.class);
    expect(heuristicFunctionFactory.createHeuristicFunction(heuristic))
      .andReturn(heuristicFunction);

    Route route=createMock(DefaultRoute.class);

    replay(costFunctionFactory, heuristicFunctionFactory);

    // Run
    RoutingServiceImpl service=createInstance(startLocation, goalLocation,
      costFunction, heuristicFunction, route);
    service.setCostFunctionFactory(costFunctionFactory);
    service.setHeuristicFunctionFactory(heuristicFunctionFactory);
    Route actualRoute=service.findRoute(startLocation, goalLocation, cost,
      heuristic);

    // Verify
    verify(costFunctionFactory, heuristicFunctionFactory);
    assertSame(route, actualRoute);
  }

  @Test(expected=NotEnabledException.class)
  public void testFindRoute2Disabled() throws Exception {
    RoutingServiceImpl service=createInstance();
    service.setEnabled(false);
    service.findRoute(null, null);
  }

  @Test
  public void testCreateSearchAlgorithm() {
    RoutingServiceImpl service=createInstance();
    assertNotNull(service.createSearchAlgorithm());
  }

  @Test
  public void testFindRouteInternal() throws Exception {
    // Init
    Coordinate location1=createMock(Coordinate.class);
    Coordinate location2=createMock(Coordinate.class);
    StreetNode node=createMock(StreetNode.class);
    CostFunction costFunction=createMock(CostFunction.class);
    HeuristicFunction heuristicFunction=createMock(HeuristicFunction.class);
    FindingDao findingDao=createMock(FindingDao.class);
    Route route=createMock(Route.class);
    int timeout=(int)(Math.random()*100000+1);

    DefaultAStarSearchAlgorithm searchAlg=createMock(DefaultAStarSearchAlgorithm.class);
    searchAlg.setStartNode(node);
    searchAlg.setGoalNode(node);
    searchAlg.setCostFunction(costFunction);
    searchAlg.setHeuristicFunction(heuristicFunction);
    searchAlg.setFindingDao(findingDao);
    searchAlg.setTimeout(timeout);
    expect(searchAlg.run()).andReturn(route);

    replay(searchAlg);

    // Run
    RoutingServiceImpl service=createInstance(location1, location2, node,
      searchAlg);
    service.setFindingDao(findingDao);
    service.setTimeout(timeout);
    Route actualRoute=service.findRoute(location1, location2, costFunction,
      heuristicFunction);

    // Verify
    verify(searchAlg);
    assertSame(route, actualRoute);
  }

  @Test
  public void testFindNearestStreetNode() {
    // Init
    StreetNode node=createMock(StreetNode.class);
    Coordinate c=new Coordinate(15, 46);
    GeometryFactory gf=new GeometryFactory();
    Point point=gf.createPoint(c);

    GeometryFactory gfMock=createMock(GeometryFactory.class);
    expect(gfMock.createPoint(c)).andReturn(point);

    FindingDao findingDao=createMock(FindingDao.class);
    expect(findingDao.findNearestStreetNode(point, 10000)).andReturn(node);

    replay(gfMock, findingDao);

    // Run
    RoutingServiceImpl service=createInstance();
    service.setGeometryFactory(gfMock);
    service.setFindingDao(findingDao);
    StreetNode actualNode=service.findNearestStreetNode(c);

    // Verify
    verify(gfMock, findingDao);
    assertSame(node, actualNode);
  }
}
